;created as an attempt to replicate results found by Cremers as expressed in Key Exchange in IPsec Revisited.
;(Or rather from input of ikev1-pk2-a.cpp)
(herald "IKEv1 Public-Key" (algebra diffie-hellman) (bound 15))


;Macros to reflect what certian parties do not know
(defmacro (zi) (exp Gr i))
(defmacro (zr) (exp Gi r))


;Macros for the different messages sent back and forth
(defmacro (hash_Ii) (hash ni nr (zi) (exp (gen) i) Gr ci cr lst I))
(defmacro (hash_Ir) (hash ni nr (zr) Gi (exp (gen) r) ci cr lst I))
(defmacro (hash_Ri) (hash ni nr (zi) (exp (gen) i) Gr cr ci lst R))
(defmacro (hash_Rr) (hash ni nr (zr) Gi (exp (gen) r) cr ci lst R))
(defmacro (ilength) 2)
(defmacro (rlength) 3)

(defmacro (SKi) (hash "KDF" ni nr (zi) ci cr))
(defmacro (SKr) (hash "KDF" ni nr (zr) ci cr))
(defmacro (SK)  (hash "KDF" ni nr (exp (gen) (mul i r)) ci cr))

(defmacro (Nei) (hash "prf" ni ci))
(defmacro (Ner) (hash "prf" nr cr))

(defprotocol pk2-a diffie-hellman
	     (defrole init
		      (vars (ni nr ci cr lst algo text) (i expn) (Gr base) (I R name))
		      (trace
			(send (cat I R ci lst (enc ni (pubk R)) (enc (exp (gen) i) (Nei)) (enc I (Nei))))
			(recv (cat R I ci cr algo (enc nr (pubk I)) (enc Gr (Ner)) (enc R (Ner)) (hash_Ri)))
			(send (cat I R ci cr (hash_Ii))))
		      (uniq-orig ni ci)
		      (uniq-gen i)
                      (neq (Gr (gen))))
	      (defrole resp
		     (vars (ni nr ci cr lst algo text) (r expn) (Gi base) (I R name))
		     (trace
		       (recv (cat I R ci lst (enc ni (pubk R)) (enc Gi (Nei)) (enc I (Nei))))
		       (send (cat R I ci cr algo (enc nr (pubk I)) (enc (exp (gen) r) (Ner)) (enc R (Ner)) (hash_Rr)))
		       (recv (cat I R ci cr (hash_Ir)))
                       (send "blab"))
		       (uniq-orig nr cr)
		      (uniq-gen r)
                      (neq (Gi (gen))))
              (defrole blabber-akey
               (vars (key akey))
               (trace
                (recv "blab")
                (send key))
               (uniq-orig key)))

;;;
;;; Authentication skeletons
;;;

;--------------------------------

(defgoal pk2-a
  ; Aliveness - my peer has taken an action
  (forall ((iend brend brstart node) (I R name) (ni nr ci cr lst algo
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(p "blabber-akey" 1 brend) (str-prec brstart brend)
		(p "blabber-akey" 0 brstart) (p "blabber-akey" "key" brend (privk R))
		(prec iend brstart)
		(pnon (privk R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((iend brend brstart node) (I R name) (ni nr ci cr lst algo
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(p "blabber-akey" 1 brend) (str-prec brstart brend)
		(p "blabber-akey" 0 brstart) (p "blabber-akey" "key" brend (privk R))
		(prec iend brstart)
		(pnon (privk R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R) (p "init" "R" i1 I))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((iend brend brstart node) (I R name) (ni nr ci cr lst algo
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(p "blabber-akey" 1 brend) (str-prec brstart brend)
		(p "blabber-akey" 0 brstart) (p "blabber-akey" "key" brend (privk R))
		(prec iend brstart)
		(pnon (privk R)))
	   (exists ((r1 node) (Gi base) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "Gi" r1 Gi) (p "resp" "r" r1 r)
			(= (exp Gi r) (exp Gr i))))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((iend brend brstart node) (I R name) (ni nr ci cr lst algo
                  text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(p "blabber-akey" 1 brend) (str-prec brstart brend)
		(p "blabber-akey" 0 brstart) (p "blabber-akey" "key" brend (privk R))
		(prec iend brstart)
		(pnon (privk R)))
	   (exists ((r1 node) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "ni" r1 ni) (p "resp" "nr" r1 nr)
			(p "resp" "r" r1 r) (p "resp" "ci" r1 ci)
			(p "resp" "cr" r1 cr) (p "resp" "lst" r1 lst)
			(p "resp" "Gi" r1 (exp (gen) i))
			(p "resp" "algo" r1 algo)))))
  )

(defgoal pk2-a
  ; Aliveness - my peer has taken an action
  (forall ((rend biend bistart node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(p "blabber-akey" 1 biend) (str-prec bistart biend)
		(p "blabber-akey" 0 bistart) (p "blabber-akey" "key" biend (privk I))
		(prec rend bistart)
		(pnon (privk I)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((rend biend bistart node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(p "blabber-akey" 1 biend) (str-prec bistart biend)
		(p "blabber-akey" 0 bistart) (p "blabber-akey" "key" biend (privk I))
		(prec rend bistart)
		(pnon (privk I)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I) (p "resp" "I" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I) (p "init" "R" i1 R))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((rend biend bistart node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(p "blabber-akey" 1 biend) (str-prec bistart biend)
		(p "blabber-akey" 0 bistart) (p "blabber-akey" "key" biend (privk I))
		(prec rend bistart)
		(pnon (privk I)))
	   (exists ((i1 node) (Gr base) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
		        (p "init" "Gr" i1 Gr) (p "init" "i" i1 i)
			(= (exp Gi r) (exp Gr i))
			))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((rend biend bistart node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(p "blabber-akey" 1 biend) (str-prec bistart biend)
		(p "blabber-akey" 0 bistart) (p "blabber-akey" "key" biend (privk I))
		(prec rend bistart)
		(pnon (privk I)))
	   (exists ((i1 node) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
			(p "init" "ni" i1 ni) (p "init" "nr" i1 nr)
			(p "init" "ci" i1 ci) (p "init" "cr" i1 cr)
			(p "init" "algo" i1 algo) (p "init" "lst" i1 lst)
		        (p "init" "Gr" i1 (exp (gen) r)) (p "init" "i" i1 i)
			))))
  (comment "Responder point of view; agreement / weak agreement / aliveness goals")
  )

;;;
;;; Key secrecy skeletons (no privk leaks)
;;;

(defskeleton pk2-a
	    (vars (ni nr ci cr lst algo text) (i expn) (I R name) (Gr base))
	    (defstrand init 3 (I I) (R R) (i i) (ni ni) (nr nr) (ci ci) (cr cr) (Gr Gr))
            (deflistener (SKi))
	    (non-orig (privk I) (privk R)))

(defskeleton pk2-a
	    (vars (ni nr ci cr lst algo text) (r expn) (I R name) (Gi base))
	    (defstrand resp 3 (I I) (R R) (ci ci) (ni ni) (nr nr) (cr cr) (Gi Gi) (r r))
            (deflistener (SKr))
	    (non-orig (privk I) (privk R)))

;;;
;;; Forward secrecy skeletons
;;;

;;; Weak Forward Secrecy
(defskeleton pk2-a
  (vars (ci lst algo ni nr cr text) (I R name) (i r expn))
  (defstrand resp 4 (ni ni) (nr nr) (ci ci) (cr cr) (lst lst)
    (algo algo) (I I) (R R) (Gi (exp (gen) i)) (r r))
  (defstrand init 3 (ni ni) (nr nr) (ci ci) (cr cr) (lst lst)
    (algo algo) (I I) (R R) (Gr (exp (gen) r)) (i i))
  (defstrand blabber-akey 2 (key (privk I)))
  (defstrand blabber-akey 2 (key (privk R)))
  (deflistener (SK))
  (precedes ((0 1) (1 1)) ((1 0) (0 0)) ((1 2) (0 2)) ((0 3) (2 0)) ((0 3) (3 0)))
  (ind-zero-in (r (exp (gen) i)))
  (neq ((exp (gen) i) (gen)) ((exp (gen) r) (gen)))
  (pen-non-orig (privk I) (privk R))
  (uniq-gen i r)
  (uniq-orig ci ni nr cr))


;;; Perfect Forward Secrecy
(defskeleton pk2-a
	    (vars (ni nr ci cr lst algo text) (i expn) (I R name) (Gr base))
	    (defstrand init 3 (I I) (R R) (i i) (ci ci) (cr cr) (Gr Gr) (ni ni) (nr nr))
            (defstrand blabber-akey 2 (key (privk I)))
            (defstrand blabber-akey 2 (key (privk R)))
            (deflistener (SKi))
            (precedes ((0 2) (1 0)) ((0 2) (2 0)))
	    (pen-non-orig (privk I) (privk R)))

(defskeleton pk2-a
	    (vars (ni nr ci cr lst algo text) (r expn) (I R name) (Gi base))
	    (defstrand resp 4 (I I) (R R) (ci ci) (cr cr) (r r) (Gi Gi) (ni ni) (nr nr))
            (defstrand blabber-akey 2 (key (privk I)))
            (defstrand blabber-akey 2 (key (privk R)))
            (deflistener (SKr))
            (precedes ((0 3) (1 0)) ((0 3) (2 0)))
	    (pen-non-orig (privk I) (privk R)))
