;created as an attempt to replicate results found by Cremers as expressed in Key Exchange in IPsec Revisited.
;(Or rather from input of ikev1-psk-m.cpp)
(herald "IKEv1 Pre-Shared-Key" (algebra diffie-hellman) (bound 15))


;Macros to reflect what certian parties do not know
(defmacro (zi) (exp Gr i))
(defmacro (zr) (exp Gi r))

;Macros for the different messages sent back and forth
(defmacro (hash_Ii) (hash (ltk I R) ni nr (exp (gen) i) Gr ci cr lst I))
(defmacro (hash_Iii) (hash (ltk R I) ni nr (exp (gen) i) Gr ci cr lst I))

(defmacro (hash_Ir) (hash (ltk I R) ni nr Gi (exp (gen) r) ci cr lst I))
(defmacro (hash_Irr) (hash (ltk R I) ni nr Gi (exp (gen) r) ci cr lst I))

(defmacro (hash_Ri) (hash (ltk I R) ni nr Gr (exp (gen) i) cr ci lst R))
(defmacro (hash_Rii) (hash (ltk R I) ni nr Gr (exp (gen) i) cr ci lst R))

(defmacro (hash_Rr) (hash (ltk I R) ni nr (exp (gen) r) Gi cr ci lst R))
(defmacro (hash_Rrr) (hash (ltk R I) ni nr (exp (gen) r) Gi cr ci lst R))

(defmacro (ilength) 2)
(defmacro (rlength) 3)

(defmacro (SKi) (hash "prf" (ltk I R) ni nr (zi) ci cr))
(defmacro (SKr) (hash "prf" (ltk I R) ni nr (zr) ci cr))

(defmacro (SKf1) (hash "prf" (ltk I R) ni nr (exp (gen) (mul i r)) ci cr))
(defmacro (SKf2) (hash "prf" (ltk R I) ni nr (exp (gen) (mul i r)) ci cr))

(defprotocol pskAggr diffie-hellman
	     (defrole init
		      (vars (ni nr ci cr lst algo text) (i expn) (Gr base) (I R name))
		      (trace
			(send (cat I R ci lst (exp (gen) i) ni I))
			(recv (cat R I ci cr algo Gr nr R (hash_Ri) (hash_Rii)))
			(send (cat I R ci cr (hash_Ii) (hash_Iii))))
		      (uniq-orig ni ci)
		      (uniq-gen i)
                      (neq (Gr (gen))))
	     
	     (defrole resp
		     (vars (ni nr ci cr lst algo text) (r expn) (Gi base) (I R name))
		     (trace
			(recv (cat I R ci lst Gi ni I))
			(send (cat R I ci cr algo (exp (gen) r) nr R (hash_Rr) (hash_Rrr)))
			(recv (cat I R ci cr (hash_Ir) (hash_Irr)))
                        (send "blab"))
		      (uniq-orig nr cr)
		      (uniq-gen r)
                      (neq (Gi (gen))))
             (defrole blabber-skey
               (vars (key skey))
               (trace
                (recv "blab")
                (send key))
               (uniq-orig key)))

;;;
;;; Authentication skeletons
;;;


;--------------------------------

(defgoal pskAggr
  ; Aliveness - my peer has taken an action
  (forall ((iend node) (I R name) (ni nr ci cr lst algo text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((iend node) (I R name) (ni nr ci cr lst algo text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 R) (p "init" "R" i1 I))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((iend node) (I R name) (ni nr ci cr lst algo text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((r1 node) (Gi base) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "Gi" r1 Gi) (p "resp" "r" r1 r)
			(= (exp Gi r) (exp Gr i))))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((iend node) (I R name) (ni nr ci cr lst algo text) (i expn) (Gr base))
	  (implies
	   (and (p "init" (ilength) iend) 
		(p "init" "R" iend R) (p "init" "I" iend I)
		(p "init" "ni" iend ni)
		(p "init" "nr" iend nr)
		(p "init" "ci" iend ci)
		(p "init" "cr" iend cr)
		(p "init" "lst" iend lst)
		(p "init" "i" iend i)
		(p "init" "Gr" iend Gr)
		(p "init" "algo" iend algo)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((r1 node) (r expn))
		   (and (p "resp" 0 r1) (p "resp" "R" r1 R) (p "resp" "I" r1 I)
			(p "resp" "ni" r1 ni) (p "resp" "nr" r1 nr)
			(p "resp" "r" r1 r) (p "resp" "ci" r1 ci)
			(p "resp" "cr" r1 cr) (p "resp" "lst" r1 lst)
			(p "resp" "Gi" r1 (exp (gen) i))
			(p "resp" "algo" r1 algo)))))
  )

(defgoal pskAggr
  ; Aliveness - my peer has taken an action
  (forall ((rend node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I))))))
  ; Weak agreement - my peer and I agree on our names
  (forall ((rend node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(non (ltk R I)) (non (ltk I R)))
	   (or (exists ((r1 node))
		       (and (p "resp" 0 r1) (p "resp" "R" r1 I) (p "resp" "I" r1 R)))
	       (exists ((i1 node))
		       (and (p "init" 0 i1) (p "init" "I" i1 I) (p "init" "R" i1 R))))))
  ; Minimal agreement - my peer and I agree on our names and roles, and on the DH key
  (forall ((rend node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((i1 node) (Gr base) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
		        (p "init" "Gr" i1 Gr) (p "init" "i" i1 i)
			(= (exp Gi r) (exp Gr i))
			))))
  ; Full agreement - my peer and I agree on the full transcript
  (forall ((rend node) (I R name) (ni nr ci cr lst algo text) (r expn) (Gi base))
	  (implies
	   (and (p "resp" (rlength) rend) 
		(p "resp" "R" rend R) (p "resp" "I" rend I) (p "resp" "ni" rend ni)
		(p "resp" "nr" rend nr) (p "resp" "ci" rend ci) (p "resp" "cr" rend cr)
		(p "resp" "lst" rend lst) (p "resp" "algo" rend algo) (p "resp" "r" rend r)
		(p "resp" "Gi" rend Gi)
		(non (ltk R I)) (non (ltk I R)))
	   (exists ((i1 node) (i expn))
		   (and (p "init" 0 i1) (p "init" "R" i1 R) (p "init" "I" i1 I)
			(p "init" "ni" i1 ni) (p "init" "nr" i1 nr)
			(p "init" "ci" i1 ci) (p "init" "cr" i1 cr)
			(p "init" "algo" i1 algo) (p "init" "lst" i1 lst)
		        (p "init" "Gr" i1 (exp (gen) r)) (p "init" "i" i1 i)
			))))
  (comment "Responder point of view; agreement / weak agreement / aliveness goals")
  )

;;;
;;; Key secrecy skeletons (no privk leaks)
;;;

(defskeleton pskAggr
	    (vars (ni nr ci cr lst algo text) (i expn) (Gr base) (I R name))
	    (defstrand init 3 (I I) (R R) (Gr Gr) (i i) (ni ni) (nr nr) (ci ci) (cr cr))
            (deflistener (SKi))
	    (non-orig (ltk I R) (ltk R I)))

(defskeleton pskAggr
	    (vars (ni nr ci cr lst algo text) (r expn) (Gi base) (I R name))
	    (defstrand resp 3 (I I) (R R) (Gi Gi) (r r) (ci ci) (ni ni) (nr nr) (cr cr))
            (deflistener (SKr))
	    (non-orig (ltk I R) (ltk R I)))



;;;
;;; Forward Secrecy skeletons (privk leaks)
;;;

;;; Weak Forward Secrecy
(defskeleton pskAggr
  (vars (ci lst algo ni nr cr text) (I R name) (r i expn))
  (defstrand resp 4 (ni ni) (nr nr) (ci ci) (cr cr) (lst lst)
    (algo algo) (I I) (R R) (Gi (exp (gen) i)) (r r))
  (defstrand init 3 (ni ni) (nr nr) (ci ci) (cr cr) (lst lst)
    (algo algo) (I I) (R R) (Gr (exp (gen) r)) (i i))
  (defstrand blabber-skey 2 (key (ltk I R)))
  (defstrand blabber-skey 2 (key (ltk R I)))
  (deflistener (SKf1))
  (precedes ((0 1) (1 1)) ((1 0) (0 0)) ((1 2) (0 2)) ((0 3) (2 0)) ((0 3) (3 0)))
  (ind-zero-in (r (exp (gen) i)))
  (neq ((exp (gen) r) (gen)) ((exp (gen) i) (gen)))
  (pen-non-orig (ltk I R) (ltk R I))
  (uniq-gen r i)
  (uniq-orig ci ni nr cr))

(defskeleton pskAggr
  (vars (ci lst algo ni nr cr text) (I R name) (r i expn))
  (defstrand resp 4 (ni ni) (nr nr) (ci ci) (cr cr) (lst lst)
    (algo algo) (I I) (R R) (Gi (exp (gen) i)) (r r))
  (defstrand init 3 (ni ni) (nr nr) (ci ci) (cr cr) (lst lst)
    (algo algo) (I I) (R R) (Gr (exp (gen) r)) (i i))
  (defstrand blabber-skey 2 (key (ltk I R)))
  (defstrand blabber-skey 2 (key (ltk R I)))
  (deflistener (SKf2))
  (precedes ((0 1) (1 1)) ((1 0) (0 0)) ((1 2) (0 2)) ((0 3) (2 0)) ((0 3) (3 0)))
  (ind-zero-in (r (exp (gen) i)))
  (neq ((exp (gen) r) (gen)) ((exp (gen) i) (gen)))
  (pen-non-orig (ltk I R) (ltk R I))
  (uniq-gen r i)
  (uniq-orig ci ni nr cr))

;;; Perfect Forward Secrecy
(defskeleton pskAggr
	    (vars (ni nr ci cr lst algo text) (i expn) (Gr base) (I R name))
	    (defstrand init 3 (I I) (R R) (Gr Gr) (i i) (ni ni) (nr nr) (ci ci) (cr cr))
            (defstrand blabber-skey 2 (key (ltk I R)))
            (defstrand blabber-skey 2 (key (ltk R I)))
            (deflistener (SKi))
            (precedes ((0 2) (1 0)) ((0 2) (2 0)))
	    (pen-non-orig (ltk I R) (ltk R I)))

(defskeleton pskAggr
	    (vars (ni nr ci cr lst algo text) (r expn) (Gi base) (I R name))
	    (defstrand resp 4 (I I) (R R) (Gi Gi) (r r) (ni ni) (nr nr) (ci ci) (cr cr))
            (defstrand blabber-skey 2 (key (ltk I R)))
            (defstrand blabber-skey 2 (key (ltk R I)))
            (deflistener (SKr))
            (precedes ((0 3) (1 0)) ((0 3) (2 0)))
	    (pen-non-orig (ltk I R) (ltk R I)))






